#include <sys/errno.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/ucred.h>

#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "transaction.h"
#include "util.h"
#include "logic.h"
#include "fsevent.h"

int eventfd;	// File system events come from here

static void process_command(int fd);

int
main(int argc, char *argv[])
{
	int listenfd;	// Sockets open here
	int cmdfd;	// Commands come here
	struct sockaddr_un servaddr;
	int flags;

	/* Verify startup conditions. */
	if (argc != 1)
		errx(1, "no arguments are supported");
	if (geteuid() != 0)
		errx(1, "only execution by root is allowed");

	/* Setup command listener. */
	(void)unlink(CMD_FILE);
	if ((listenfd = socket(AF_LOCAL, SOCK_STREAM, 0)) == -1)
		err(1, "socket");
	if (listenfd <= 2)
		errx("called with fd 0-2 closed; daemon(3) will fail");
	/*
	 * Set non-blocking, to avoid race condition; see Stevens et al
	 * Unix Network Programming, 3rd ed. 16.6 (p. 463).
	 */
	if ((flags = fcntl(listenfd, F_GETFL, 0)) < 0)
		err(1, "fcntl F_GETFL");
	flags |= O_NONBLOCK;
	if (fcntl(listenfd, F_SETFL, flags) < 0)
		err(1, "fcntl F_SETFL");

	/* Listen for commands from the control process.. */
	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sun_family = AF_LOCAL;
	strcpy(servaddr.sun_path, CMD_FILE);
	umask(0111);
	if (bind(listenfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) == -1)
		err(1, "bind");
	if (listen(listenfd, 1024) == -1)
		err(1, "listen");
	umask(0022);
#ifdef DEBUG
	openlog(argv[0], LOG_PERROR, LOG_LOCAL1);
#else
	openlog(argv[0], LOG_PID, LOG_LOCAL1);
	#ifndef NO_DAEMON
	daemon(0, 0);
	#endif
#endif
	syslog(LOG_INFO, "started");
	eventfd = -1;
	/* Accept commands and system logging events. */
	for (;;) {
		fd_set rset;
		struct timeval timeout;

		FD_ZERO(&rset);
		if (eventfd != -1)
			FD_SET(eventfd, &rset);
		FD_SET(listenfd, &rset);
		#ifdef DEBUG_DROPPED_EVENTS
		syslog(LOG_DEBUG, "calling select eventfd=%d", eventfd);
		#endif
		/*
		 * Under Darwin 9.6.0 and heavy load (e.g. untarring Linux)
		 * select sans timeout often gets stuck resulting in
		 * dropped events.  The timeout prevents this.
		 *
		 * Jan 12 15:42:33: calling select eventfd=8
		 * Jan 12 15:42:33: select returns
		 * Jan 12 15:42:33: read 2238 bytes
		 * Jan 12 15:42:33: calling select eventfd=8
		 * Jan 12 15:43:08: select returns
		 * Jan 12 15:43:08: read 563277 bytes
		 */
		timeout.tv_sec = 0;
		timeout.tv_usec = 10000;	/* 10ms */
		switch (select(max(eventfd, listenfd) + 1, &rset, NULL, NULL, eventfd != -1 ? &timeout : NULL)) {
		case -1:		/* Error */
			switch (errno) {
			case EINTR:
			case EAGAIN:
				continue;
			default:
				syslog(LOG_ERR, "select: %w");
				exit(1);
			}
		case 0:			/* Timeout */
			continue;
		}
		#ifdef DEBUG_DROPPED_EVENTS
		syslog(LOG_DEBUG, "select returns with data");
		#endif
		if (eventfd != -1 && FD_ISSET(eventfd, &rset))
			fsevent_read();
		if (FD_ISSET(listenfd, &rset))
			process_command(listenfd);
	} /* NOTREACHED */
}

/* Process a command from the user control process. */
static void
process_command(int listenfd)
{
	int connfd, errfd;
	struct sockaddr_un cliaddr;
	socklen_t clilen;
	struct s_command cmd;
	uid_t euid;
	gid_t egid;

	if ((connfd = accept(listenfd, (struct sockaddr *)&cliaddr, &clilen)) < 0)
		switch (errno) {
			case EWOULDBLOCK:	/* Berkeley */
			case ECONNABORTED:	/* POSIX */
			case EPROTO:		/* SVR4 */
			case EINTR:
				return;
			default:
				syslog(LOG_ERR, "accept: %w");
				exit(1);
		}

	if (getpeereid(connfd, &euid, &egid) < 0) {
		syslog(LOG_ERR, "Invalid credential data");
		close(connfd);
		return;
	}

	/* Read the actual command. */
	if (readn(connfd, &cmd, sizeof(cmd)) == -1) {
		close(connfd);
		return;
	}

	syslog(LOG_INFO, "command: '%c' from user %d", cmd.cmd, euid);
	switch (cmd.cmd) {
	case 'b':
		errfd = receive_fd(connfd);
		transaction_begin(connfd, euid, cmd.n, errfd);
		break;
	case 'c':
		transaction_commit(connfd, euid, cmd.n);
		break;
	case 'a':
		transaction_abort(connfd, euid, cmd.n);
		break;
	case 'A':
		transaction_abort_all(euid, "user initiated forced abort");
		break;
	case 'l':
		transaction_list(connfd, euid, cmd.n);
		break;
	case 'x':
		transaction_abort_all(euid, "user initiated forced abort");
		syslog(LOG_INFO, "exiting");
		exit(0);
		break;
	default:
		syslog(LOG_ERR, "invalid command: '%c'", cmd.cmd);
		break;
	}
	/*
	 * By default the system will return immediately, and still try to
	 * deliver any remaining data. See Stevens 3rd ed. Sec. 7.5 (p. 202).
	 */
	if (close(connfd) < 0)
		syslog(LOG_ERR, "close: %w");
}
